;***************************************************************************
; ECEN4002/5002 DSP Lab
; Fall 2002
;
; Lab #6:  Short-time Fourier Transform Skeleton
;
; Code to perform windowed FFT on input signal (50% overlap) and
; overlap-add of IFFT.
;
; User code can be inserted between FFT and IFFT to modify spectral
; data prior to IFFT and resynthesis.
;
; Processing is organized for FFT of length FFTLEN, with I/O buffers made
; up of three FFTLEN/2 length blocks, 0,1,2.   While block 0 is being
; filled by the interrupt service routine, blocks 1:2 are multiplied
; by a raised cosine (hanning) window and sent through the FFT/IFFT.  Output
; is overlap-added to the output blocks 1:2.  Once block 0 is full of
; new input data, FFT is done on blocks 2:0 while block 1 is being filled.
; Finally, FFT is done on blocks 0:1 while block 2 is filled.  Then
; process continues repeatedly.
;
; Window function as defined by 'wingen' macro is scaled by 1/FFTLEN to
; avoid overflow within FFT.
;
;***************************************************************************
      nolist
      include 'ioequ.asm'
      include 'intequ.asm'
      include 'ada_equ.asm'
      include 'vectors.asm'  
      list
  
FFTLEN		equ	256
TBIT		EQU	@CVI(0.5+@LOG(@CVF(FFTLEN))/@LOG(2.))
BUFMOD		equ	((3*FFTLEN/2)-1)

CTRL_WD_12      equ     MIN_LEFT_ATTN+MIN_RIGHT_ATTN+LIN2+RIN2
CTRL_WD_34      equ     MIN_LEFT_GAIN+MIN_RIGHT_GAIN

; Declare left and right input and output buffers.
; Also declare variables to hold input/output address
; pointers for use by the interrupt service routines.
       org    x:0
LINBUFFER	dsm	3*FFTLEN/2
RINBUFFER	dsm	3*FFTLEN/2
LOUTBUFFER	dsm	3*FFTLEN/2
ROUTBUFFER	dsm	3*FFTLEN/2
LEFT_IN		ds	1
RIGHT_IN	ds	1
LEFT_OUT	ds	1
RIGHT_OUT	ds	1

; Declare the FFT input, output, and coefficient arrays.
; These are declared in "L:" memory as complex numbers: Real in X,
; Imaginary in Y
	org	l:$1000
fft_in		dsm	FFTLEN
fft_out		dsm	FFTLEN
fft_out2	dsm	FFTLEN
coef		dsm	FFTLEN  ; forward FFT twiddles (filled in below)
icoef		dsm	FFTLEN  ; inverse FFT twiddles (filled in below)

	org	x:
window	dsm	FFTLEN          ; window function (filled in below)

	org	y:
tstack		ds	$80     ; Reserve some space for save/restore stack


      	include 'sincosw.asm'	    ;macro to...
        sincos  FFTLEN,coef,icoef   ;...build twiddle factor tables
        wingen  FFTLEN,window       ; ... and generate window function 


	include 'fftr2cn.asm'    ; Code for FFT macros

;******************************************************************************
;Main Program
;******************************************************************************

        org     p:$400
START
        movep   #$040007,x:M_PCTL  ; PLL 8 X 12.288 = 98.304MHz
        movep   #$012421,x:M_BCR   ; AARx - 1 wait state
        ori     #3,mr              ; mask interrupts
        movec   #0,sp              ; clear hardware stack pointer
        move    #0,omr             ; operating mode 0

; initialize the I/O buffer address pointers
	move	#LINBUFFER,r0
	move	r0,x:LEFT_IN
	move	#LOUTBUFFER,r0
	move	r0,x:LEFT_OUT
	move	#RINBUFFER,r0
	move	r0,x:RIGHT_IN
	move	#ROUTBUFFER,r0
	move	r0,x:RIGHT_OUT


; Pointers ready, so initialize the codec, and start interrupts
        jsr     codec_init       

; At this point interrupts are occurring and data is filling
;  in block 0.  

; Set transfer position to block 2 (block "before" new data)
;
; R registers set as:
; r0 LINBUFFER   m0 BUFMOD      n0 FFTLEN/2
; r1 LOUTBUFFER  m1 BUFMOD      n1 FFTLEN/2
; r2 RINBUFFER   m2 BUFMOD
; r3 ROUTBUFFER  m3 BUFMOD
; r4 window      m4 FFTLEN-1
; r5 fft_in      m5 FFTLEN-1
; r6 fft_out     m6 FFTLEN-1    n6 FFTLEN/2
; r7 save stack  m7 -1 (linear)
 
	move	#LINBUFFER+FFTLEN,r0
	move	#LOUTBUFFER+FFTLEN,r1
	move	#RINBUFFER+FFTLEN,r2
	move	#ROUTBUFFER+FFTLEN,r3
	move	#BUFMOD,m0
	move	m0,m1
	move	m0,m2
	move	m0,m3

	move	#window,r4
	move	#fft_in,r5
	move 	#fft_out,r6
	move	#(FFTLEN-1),m4
	move	m4,m5
	move 	m4,m6
	move 	#FFTLEN/2,n0
	move 	n0,n1
	move 	n0,n6

	move	#tstack,r7
	move	#-1,m7
;***************************************************************************
; Begin process loop...
;  Calls to 'do_stft_process' apply FFT/IFFT to LEFT channel, and just
;  copy RIGHT channel input to the output.
;***************************************************************************
loop

; Look for address bit being set (00-> 01): means block 0 has been read.
waitb0
	move	x:LEFT_IN,x0
	jclr	#TBIT-1,x0,waitb0

	jsr	do_stft_process

; Now look for address bit going clear (01-> 10):  block 1 has been read.
waitb1
	move	x:LEFT_IN,x0
	jset	#TBIT-1,x0,waitb1

	jsr	do_stft_process

; Now look for next address bit going clear (10-> 00): block 2 read.
waitb2
	move	x:LEFT_IN,x0
	jset	#TBIT,x0,waitb2

	jsr	do_stft_process

        jmp     loop   ; continue with I/O loop


;***************************************************************************
; do_stft_process
;
; Subroutine for windowed overlap-add short-time Fourier transform
; using the FFT.  Code currently written to process LEFT channel ONLY,
; and just copy right channel data from input to output.
;***************************************************************************

do_stft_process
; Handle left channel only for now, so just copy right channel
; data from input buffer (r2) to output buffer (r3)
	do #FFTLEN/2,rcpy
	move x:(r2)+,x0
	move x0,x:(r3)+
rcpy    nop

; Get data from left input, mult by window, and store in real part
; of fft_in, store zeros in fft_in .
	clr b
	do #FFTLEN,wincp
        move x:(r0)+,x0	     ; read starts in block n, then thru block n+1 
	move x:(r4)+,y0      ; window function in r4 
	mpy x0,y0,a          ; apply window
	move b,y:(r5)        ; fft_in in r5
	move a,x:(r5)+
wincp

; FFT input ready.
; FFT routine uses essentially all the registers, so NEED to 
; save [R,M,N] registers here.  Stack is via r7.
	move r0,y:(r7)+
	move r1,y:(r7)+
	move r2,y:(r7)+
	move r3,y:(r7)+
	move r4,y:(r7)+
	move r5,y:(r7)+
	move r6,y:(r7)+
	move m0,y:(r7)+
	move m1,y:(r7)+
	move m2,y:(r7)+
	move m3,y:(r7)+
	move m4,y:(r7)+
	move m5,y:(r7)+
	move m6,y:(r7)+
	move n0,y:(r7)+
	move n1,y:(r7)+
	move n2,y:(r7)+
	move n3,y:(r7)+
	move n4,y:(r7)+
	move n5,y:(r7)+
	move n6,y:(r7)+

; Do FFT on l:fft_in
	fftr2cn   FFTLEN,fft_in,fft_out2,coef
; FFT result in l:fft_out2

;***************************************************************************
; USER CODE GOES HERE!
; Your spectral processing code to manipulate spectral data in l:fft_out2 
; gets inserted here.
;
; Remember to treat complex FFT data in conjugate symmetric
; form if a real output signal is desired.
;***************************************************************************

; Do IFFT on l:fft_out2
	ifftr2cn  FFTLEN,fft_out2,fft_out,icoef
; IFFT result in l:fft_out

; NEED to restore [R,M,N] registers
	move y:-(r7),n6
	move y:-(r7),n5
	move y:-(r7),n4
	move y:-(r7),n3
	move y:-(r7),n2
	move y:-(r7),n1
	move y:-(r7),n0
	move y:-(r7),m6
	move y:-(r7),m5
	move y:-(r7),m4
	move y:-(r7),m3
	move y:-(r7),m2
	move y:-(r7),m1
	move y:-(r7),m0
	move y:-(r7),r6
	move y:-(r7),r5
	move y:-(r7),r4
	move y:-(r7),r3
	move y:-(r7),r2
	move y:-(r7),r1
	move y:-(r7),r0

; Now overlap add current block to output, and copy next block to output
; r6 has fft_out address, r1 has left output buffer address
	do #FFTLEN/2,end_ola
	move x:(r6+n6),x0      ; get block n+1 portion
	move x:(r6)+,a         ; get block n portion
	move x:(r1),y0         ; get block n prev output
	add  y0,a
	move x0,x:(r1+n1)      ; copy block n+1 portion
	move a,x:(r1)+         ; save overlap add (block n) portion
end_ola

; Update r0 (LINBUFFER) pointer to previous block
; Also update r6 (fft_out) to point to head for next time
	move	(r0)-n0
	move    (r6)+n6
	nop
	rts                    ; end of do_stft_process


	include 'cdc_init.asm'

        end

